#include <stdlib.h>
#define USE_AND_OR /* To enable AND_OR mask setting */
#include <ports.h>
#include <pps.h>
#include <timer.h>
#include <outcompare.h>
#include <adc.h>
//#include "global.h"
#include "btstack/debug.h"
#include "btstack/btstack.h"
#include <btstack/hci_cmds.h>
#include <btstack/run_loop.h>
#include <btstack/sdp_util.h>

#include "btstack/hci.h"
#include "btstack/l2cap.h"
#include "btstack/btstack_memory.h"
#include "btstack/remote_device_db.h"
#include "btstack/rfcomm.h"
#include "btstack/sdp.h"
//#include "btstack/config.h"

#define HEARTBEAT_PERIOD_MS		1000
#define	CREDITS					  10
#define INQUIRY_INTERVAL		  15

#define	led1_on_mac()			{led1_on(); led1_on_count=timer_counter; led1_on_state=1;}
//static uint8_t				rfcomm_channel_nr = 1;
//static uint16_t				rfcomm_channel_id;
//static uint8_t				spp_service_buffer[128];
static volatile clock_t			timer_counter=0;
static unsigned					led1_on_count=0;
static uint8_t					led1_on_state=0;
static char						lineBuffer[80];
static uint8_t					use_extension=0;
#if	UART1_MINIMUM_INTERVAL
static clock_t					lastsendtime=0;
#endif
#if	SEND_ON_DIFFERENTIAL || NEUTRAL_DATA_SUPPRESS
static char						lastsendbuf[8];
#endif

#define	BUTTON_STATE_INIT	0

const char *hexMap = "0123456789ABCDEF";

static bd_addr_t keyboard;
static int haveKeyboard = 0;
static WORD wii_button_state=BUTTON_STATE_INIT;
static WORD	wii_remocon_channel_id=0;
static WORD	sensor_advalue[6];

#define	ACC_INIT	0x80
static BYTE	wii_acc_xx=ACC_INIT,wii_acc_yy=ACC_INIT,wii_acc_zz=ACC_INIT;		//accelerometer value x,y,z

hci_transport_t * hci_transport_picusb_instance();

#if defined(__PIC24FJ64GB004__)
	_CONFIG1(WDTPS_PS2048 & FWPSA_PR128 & WINDIS_OFF & FWDTENDEF & ICSDEF & GWRP_OFF & GCP_OFF & JTAGEN_OFF)
	_CONFIG2(POSCMOD_NONE & I2C1SEL_PRI & IOL1WAY_OFF & OSCIOFNC_ON & FCKSM_CSDCMD & FNOSC_FRCPLL & PLL96MHZ_ON & PLLDIV_DIV2 & IESO_OFF)
	_CONFIG3(WPFP_WPFP0 & SOSCSEL_IO & WUTSEL_LEG & WPDIS_WPDIS & WPCFG_WPCFGDIS & WPEND_WPENDMEM)
	_CONFIG4(DSWDTPS_DSWDTPS3 & DSWDTOSC_LPRC & RTCOSC_SOSC & DSBOREN_OFF & DSWDTEN_OFF)
#else

    #error Cannot define configuration bits.

#endif

BYTE        deviceAddress;  // Address of the device on the USB
bd_addr_t addr_global;

void __attribute__ ((interrupt,no_auto_psv)) _OC1Interrupt(void)
{
	OC1_Clear_Intr_Status_Bit;
}

void __attribute__ ((interrupt,no_auto_psv)) _OC2Interrupt(void)
{
	OC2_Clear_Intr_Status_Bit;
}

void __attribute__ ((interrupt,no_auto_psv)) _ADC1Interrupt(void)
{
	ADC1_Clear_Intr_Status_Bit;
	sensor_advalue[0]=ADC1BUF0;
	sensor_advalue[1]=ADC1BUF1;
	sensor_advalue[2]=ADC1BUF2;
	sensor_advalue[3]=ADC1BUF3;
	sensor_advalue[4]=ADC1BUF4;
	sensor_advalue[5]=ADC1BUF5;
}

void __attribute__ ((interrupt,no_auto_psv)) _CNInterrupt(void)
{
	InputChange_Clear_Intr_Status_Bit;

	if(get_switch1()){
		led3_on();
	}else{
		led3_off();
	}

	if(get_switch2()){
		led4_on();
	}else{
		led4_off();
	}
}

BOOL InitializeSystem ( void )
{
   #if defined(__PIC24FJ64GB004__)
	    {
	        unsigned int pll_startup_counter = 600;
	        CLKDIVbits.PLLEN = 1;
	        while(pll_startup_counter--);
	    }

		AD1PCFG = 0xffff;
	    CLKDIV = 0;    // Set PLL prescaler (1:1)
    #endif

	led1_setup();
	led2_setup();
	led3_setup();
	led4_setup();
	led5_setup();

	led1_on();
	led2_on();
	led3_on();
	led4_on();
	led5_on();

	switch1_setup();
	switch2_setup();

	motor_a1_setup();
	motor_a2_setup();
	motor_pwma_setup();

	motor_b1_setup();
	motor_b2_setup();
	motor_pwmb_setup();

	OpenTimer2(T2_ON|PWM_PRESCALER,PWM_WIDTH);
	OpenOC1(OC_IDLE_CON|OC_TIMER2_SRC|OC_PWM_EDGE_ALIGN,OC_SYNC_TRIG_IN_TMR2,0,0);
	SetPriorityIntOC1(INT_PRI_4);
	EnableIntOC1;
	SetDCOC1PWM(0,0);

	OpenOC2(OC_IDLE_CON|OC_TIMER2_SRC|OC_PWM_EDGE_ALIGN,OC_SYNC_TRIG_IN_TMR2,0,0);
	SetPriorityIntOC2(INT_PRI_4);
	EnableIntOC2;
	SetDCOC2PWM(0,0);

	OpenADC10(
		ADC_MODULE_ON|ADC_IDLE_CONTINUE|ADC_FORMAT_INTG|ADC_CLK_AUTO|ADC_AUTO_SAMPLING_ON|ADC_SAMP_ON,
		ADC_VREF_AVDD_AVSS|ADC_SCAN_ON|/* ADC_INTR_6_CONV| */ ADC_INTR_4_CONV|ADC_ALT_BUF_OFF|ADC_ALT_INPUT_OFF,
		ADC_CONV_CLK_SYSTEM|ADC_SAMPLE_TIME_31|ADC_CONV_CLK_128Tcy,
		ENABLE_AN0_ANA|ENABLE_AN1_ANA|ENABLE_AN4_ANA|ENABLE_AN5_ANA/*|ENABLE_AN6_ANA|ENABLE_AN7_ANA*/,
		ADC_SCAN_AN0|ADC_SCAN_AN1|ADC_SCAN_AN4|ADC_SCAN_AN5/*|ADC_SCAN_AN6|ADC_SCAN_AN7*/
	);
	SetPriorityIntADC1(INT_PRI_4);
	EnableIntADC1;

	SetPriorityIntInputChange(INT_PRI_4);	
	EnableIntInputChange;

//	led1_off();
//	led2_off();
//	led3_off();
//	led4_off();
//	led5_off();

	u1tx_setup();
	u1rx_setup();
	u1cts_setup();
	u1rts_setup();
	sto_setup();	//for debug out

#if	ENABLE_KNIGHTRIDER
	kr_led1_setup();
	kr_led2_setup();
	kr_led3_setup();
	kr_led4_setup();
	kr_led5_setup();
	kr_led6_setup();
	kr_led7_setup();
	kr_led8_setup();
#endif

    // Init UART
    UART1Init(BAUDRATE1);

#if	__DEBUG
    UART2Init();
#endif

    return TRUE;
} // InitializeSystem

BOOL USB_ApplicationEventHandler ( BYTE address, USB_EVENT event, void *data, DWORD size )
{
    #ifdef USB_BLUETOOTH_SUPPORT_SERIAL_NUMBERS
        BYTE i;
    #endif

    // Handle specific events.
    switch (event)
    {
        case EVENT_BLUETOOTH_ATTACH:
            if (size == sizeof(BLUETOOTH_DEVICE_ID))
            {
                deviceAddress   = ((BLUETOOTH_DEVICE_ID *)data)->deviceAddress;
                log_info( "Bluetooth device attached - event, deviceAddress=%d\n",deviceAddress );
                #ifdef USB_BLUETOOTH_SUPPORT_SERIAL_NUMBERS
                    //for (i=1; i<((BLUETOOTH_DEVICE_ID *)data)->serialNumberLength; i++)
                    //{
                    //    xputc( ((BLUETOOTH_DEVICE_ID *)data)->serialNumber[i] );
                    //}
	                //xputs( "\n" );
                #endif
                return TRUE;
            }
            break;

        case EVENT_BLUETOOTH_DETACH:
            deviceAddress   = 0;
            log_info( "Bluetooth device detached - event\n" );
            return TRUE;

        case EVENT_BLUETOOTH_TX_CMD_DONE:           // The main state machine will poll the driver.
            return TRUE;

        case EVENT_BLUETOOTH_TX_ACL_DONE:           // The main state machine will poll the driver.
            return TRUE;

        case EVENT_BLUETOOTH_RX_EVENT_DONE:
			if(*(DWORD*)data){
				event_bluetooth_rxEvent_done((WORD)(*(DWORD*)data));
			}
            return TRUE;

        case EVENT_BLUETOOTH_RX_ACL_DONE:
			if(*(DWORD*)data){
				event_bluetooth_rxAcl_done((WORD)(*(DWORD*)data));
			}
            return TRUE;

        case EVENT_VBUS_REQUEST_POWER:
            // We'll let anything attach.
            return TRUE;

        case EVENT_VBUS_RELEASE_POWER:
            // We aren't keeping track of power.
            return TRUE;

        case EVENT_HUB_ATTACH:
            log_info( "\n***** USB Error - hubs are not supported *****\n" );
            return TRUE;
            break;

        case EVENT_UNSUPPORTED_DEVICE:
            log_info( "\n***** USB Error - device is not supported *****\n" );
            return TRUE;
            break;

        case EVENT_CANNOT_ENUMERATE:
            log_info( "\n***** USB Error - cannot enumerate device *****\n" );
            return TRUE;
            break;

        case EVENT_CLIENT_INIT_ERROR:
            log_info( "\n***** USB Error - client driver initialization error *****\n" );
            return TRUE;
            break;

        case EVENT_OUT_OF_MEMORY:
            log_info( "\n***** USB Error - out of heap memory *****\n" );
            return TRUE;
            break;

        case EVENT_UNSPECIFIED_ERROR:   // This should never be generated.
            log_info( "\n***** USB Error - unspecified *****\n" );
            return TRUE;
            break;

        case EVENT_SUSPEND:
        case EVENT_DETACH:
        case EVENT_RESUME:
        case EVENT_BUS_ERROR:
            return TRUE;
            break;

        default:
            break;
    }

    return FALSE;

} // USB_ApplicationEventHandler

#define	BUTTON_TWO			0x0001
#define	BUTTON_ONE			0x0002
#define	BUTTON_B			0x0004
#define	BUTTON_A			0x0008
#define	BUTTON_MINUS		0x0010
#define	BUTTON_HOME			0x0080
#define	BUTTON_LEFT			0x0100
#define	BUTTON_RIGHT		0x0200
#define	BUTTON_DOWN			0x0400
#define	BUTTON_UP			0x0800
#define	BUTTON_PLUS			0x1000


clock_t clock(void)
{
	return timer_counter;
}

#define	NEUTRAL_SHORT_BRAKE		1
#define	DEAD_TIME				1		//(*10us)

void out_button_state(const unsigned char *hid_report,uint16_t size)
{
	static BYTE	neutral_a=0,neutral_b=0;
	int	sense_yy;

	wii_button_state=(hid_report[2]<<8)|hid_report[3];
	wii_acc_xx=hid_report[4];		//acc_xx
	wii_acc_yy=hid_report[5];		//acc_yy
	wii_acc_zz=hid_report[6];		//acc_zz

	//clip_yy
	sense_yy=(((int)wii_acc_yy)-0x80);
	if(sense_yy<-0x10){
		sense_yy=-0x10;
	}else
	if(sense_yy>0x10){
		sense_yy=0x10;
	}

	//1{^ Oi
	if(wii_button_state&BUTTON_ONE){
		neutral_a=neutral_b=0;
		motor_a1_l();
		motor_a2_h();
		motor_b1_l();
		motor_b2_h();
		motor_pwma_write((PWM_DIGITAL_RUN_DUTY*((sense_yy>0)?(0x10-sense_yy):0x10))/0x10);
		motor_pwmb_write((PWM_DIGITAL_RUN_DUTY*((sense_yy<0)?(0x10+sense_yy):0x10))/0x10);
	}else

	//2{^ i
	if(wii_button_state&BUTTON_TWO){
		neutral_a=neutral_b=0;
		motor_a1_h();
		motor_a2_l();
		motor_b1_h();
		motor_b2_l();
		motor_pwma_write((PWM_DIGITAL_RUN_DUTY*((sense_yy>0)?(0x10-sense_yy):0x10))/0x10);
		motor_pwmb_write((PWM_DIGITAL_RUN_DUTY*((sense_yy<0)?(0x10+sense_yy):0x10))/0x10);
	}else

	//\{^ł̑
	if(wii_button_state&BUTTON_UP){
		neutral_a=neutral_b=0;
		motor_a1_l();
		motor_a2_h();
		motor_b1_l();
		motor_b2_h();
		motor_pwma_write(PWM_DIGITAL_RUN_DUTY);
		motor_pwmb_write(PWM_DIGITAL_RUN_DUTY);
	}else
	if(wii_button_state&BUTTON_DOWN){
		neutral_a=neutral_b=0;
		motor_a1_h();
		motor_a2_l();
		motor_b1_h();
		motor_b2_l();
		motor_pwma_write(PWM_DIGITAL_RUN_DUTY);
		motor_pwmb_write(PWM_DIGITAL_RUN_DUTY);
	}else
	if(wii_button_state&BUTTON_LEFT){
		neutral_a=neutral_b=0;
		motor_a1_h();
		motor_a2_l();
		motor_b1_l();
		motor_b2_h();
		motor_pwma_write(PWM_DIGITAL_RUN_DUTY);
		motor_pwmb_write(PWM_DIGITAL_RUN_DUTY);
	}else
	if(wii_button_state&BUTTON_RIGHT){
		neutral_a=neutral_b=0;
		motor_a1_l();
		motor_a2_h();
		motor_b1_h();
		motor_b2_l();
		motor_pwma_write(PWM_DIGITAL_RUN_DUTY);
		motor_pwmb_write(PWM_DIGITAL_RUN_DUTY);
	}else{
#if	DEAD_TIME
		if(!neutral_a){
			motor_a1_l();
			motor_a2_l();
			motor_pwma_write(PWM_WIDTH);
		}
		if(!neutral_b){
			motor_b1_l();
			motor_b2_l();
			motor_pwmb_write(PWM_WIDTH);
		}
		Delay10us(DEAD_TIME);
#endif
		neutral_a=neutral_b=1;
#if	NEUTRAL_SHORT_BRAKE
		motor_a1_h();
		motor_a2_h();
		motor_b1_h();
		motor_b2_h();
		motor_pwma_write(0);
		motor_pwmb_write(0);
#else
		motor_a1_l();
		motor_a2_l();
		motor_b1_l();
		motor_b2_l();
		motor_pwma_write(PWM_WIDTH);
		motor_pwmb_write(PWM_WIDTH);
#endif
	}
#if	0
	but_a_out_write(wii_button_state&BUTTON_A);
	but_1_out_write(wii_button_state&BUTTON_ONE);
	but_2_out_write(wii_button_state&BUTTON_TWO);
	but_b_out_write(wii_button_state&BUTTON_B);
	but_home_out_write(wii_button_state&BUTTON_HOME);
	but_plus_out_write(wii_button_state&BUTTON_PLUS);
	but_minus_out_write(wii_button_state&BUTTON_MINUS);
#endif
}

#if	WII_REMOTE_REPORT_MODE==0x30
const unsigned char neutral_hid_report[]={0xA1,0x30,0x00,0x00,};
#else
const unsigned char neutral_hid_report[]={0xA1,0x31,0x00,0x00,0x80,0x80,0x80,};
#endif

void init_button_state(void)
{
	wii_button_state=BUTTON_STATE_INIT;
	wii_acc_xx=ACC_INIT;
	wii_acc_yy=ACC_INIT;
	wii_acc_zz=ACC_INIT;
	out_button_state(neutral_hid_report,sizeof(neutral_hid_report));
}

void wii_remote_regread(uint16_t channel,uint32_t adr,uint16_t size)
{
	lineBuffer[0]=0xa2;
	lineBuffer[1]=0x17;
	lineBuffer[2]=(adr>>24)&0xff;
	lineBuffer[3]=(adr>>16)&0xff;
	lineBuffer[4]=(adr>>8)&0xff;
	lineBuffer[5]=(adr>>0)&0xff;
	lineBuffer[6]=0;
	lineBuffer[7]=size;
	l2cap_send_internal(channel,(uint8_t*)lineBuffer,8);
}

void wii_remote_regwrite1(uint16_t channel,uint32_t adr,uint8_t data)
{
	lineBuffer[0]=0xa2;
	lineBuffer[1]=0x16;
	lineBuffer[2]=(adr>>24)&0xff;
	lineBuffer[3]=(adr>>16)&0xff;
	lineBuffer[4]=(adr>>8)&0xff;
	lineBuffer[5]=(adr>>0)&0xff;
	lineBuffer[6]=1;
	lineBuffer[7]=data;
	memset(lineBuffer+8,0,15);
	l2cap_send_internal(channel,(uint8_t*)lineBuffer,7+16);
}

void wii_remote_report_mode(uint16_t channel,uint8_t tt,uint8_t mm)
{
	lineBuffer[0]=0xa2;
	lineBuffer[1]=0x12;
	lineBuffer[2]=tt;
	lineBuffer[3]=mm;
	l2cap_send_internal(channel,(uint8_t*)lineBuffer,4);
}

void wii_remote_request_status_report(uint16_t channel)
{
	lineBuffer[0]=0xa2;
	lineBuffer[1]=0x15;
	lineBuffer[2]=0x00;
	l2cap_send_internal(channel,(uint8_t*)lineBuffer,3);
}

#if	!SEND_ON_DIFFERENT_DATA
#define	REPORTMODE_TT_BIT			0x04
#else
#define	REPORTMODE_TT_BIT			0x00
#endif

void hid_process_packet(uint16_t channel,unsigned char *hid_report,uint16_t size)
{
	if(hid_report[0]!=0xa1){
		return;
	}

#if	WII_REMOTE_REPORT_MODE!=0x30
	if(hid_report[1]==0x30){
		wii_remote_report_mode(channel,REPORTMODE_TT_BIT,WII_REMOTE_REPORT_MODE);
		wii_remote_request_status_report(channel);
	}
#endif

	switch(hid_report[1]){
	case	0x20:									//status report
		if(!use_extension && hid_report[4]&0x02){						//An Extension Controller is connected
			wii_remote_regwrite1(channel,0x04a400f0,0x55);
			wii_remote_regwrite1(channel,0x04a400fb,0x00);
			use_extension=1;
		}else
		if(use_extension && !(hid_report[4]&0x02)){						//An Extension Controller is disconnected
			use_extension=0;
			wii_remote_report_mode(channel,REPORTMODE_TT_BIT,WII_REMOTE_REPORT_MODE);
		}
		break;
	case	0x21:									//read memory
		if(hid_report[5]==0x00 && hid_report[6]==0xfa){
			//extension type
			if(
				hid_report[ 7]==0xff
			&&	hid_report[ 8]==0xff
			&&	hid_report[ 9]==0xff
			&&	hid_report[10]==0xff
			&&	hid_report[11]==0xff
			&&	hid_report[12]==0xff
			){
				//ffff ffff ffff:Partially inserted -> Request Status Report
				use_extension=0;
				wii_remote_request_status_report(channel);
			}else
			if(
				((hid_report[ 7]==0x00) || (hid_report[7]==0x01))
			&&	hid_report[ 8]==0x00
			&&	hid_report[ 9]==0xa4
			&&	hid_report[10]==0x20
			&&	hid_report[11]==0x01
			&&	hid_report[12]==0x01
			){
				//0000 A420 0101:classic controller
				//0100 A420 0101:classic controller Pro
				wii_remote_report_mode(channel,REPORTMODE_TT_BIT,WII_REMOTE_REPORT_MODE_EXT);
			}
		}
		break;
	case	0x22:									//write memory
		wii_remote_regread(channel,0x04a400fa,6);	//read extension type
		break;
	case	0x30:									//0x30:Core Buttons
	case	0x31:									//0x31:Core Buttons and Accelerometer
	case	0x35:									//0x35:Core Buttons and Accelerometer with 16 Extension Bytes
		led1_on_mac();
		out_button_state(hid_report,size);
		break;
	default:
		break;
	}
	return;
}

static void l2cap_packet_handler(uint8_t packet_type, uint16_t channel, uint8_t *packet, uint16_t size)
{
	if (packet_type == HCI_EVENT_PACKET){
		switch(packet[0]){
		case L2CAP_EVENT_INCOMING_CONNECTION:
			log_info("L2CAP_EVENT_INCOMING_CONNECTION\n");
			l2cap_accept_connection_internal(channel);
			break;
		case L2CAP_EVENT_CHANNEL_OPENED:
	        if (packet[2]) {
	            log_info("Connection failed\n");
	            return;
		    }
			log_info("interrupt channel connected\n");
			wii_remocon_channel_id=channel;
 			break;
		case L2CAP_EVENT_CHANNEL_CLOSED:
			init_button_state();
			wii_remocon_channel_id=0;
			use_extension=0;
			haveKeyboard = 0;
#if	!MODE_NO_INQUIRY
			hci_send_cmd(&hci_inquiry, HCI_INQUIRY_LAP, INQUIRY_INTERVAL, 0);
#endif
			break;
		default:
			break;
		}	
	}
	if (packet_type == L2CAP_DATA_PACKET){
		hid_process_packet(channel,packet,size);
	}
}

//static uint16_t control_channel=0;
static void l2cap_control_packet_handler(uint8_t packet_type, uint16_t channel, uint8_t *packet, uint16_t size)
{
	if (packet_type == HCI_EVENT_PACKET){
		switch(packet[0]){
		case L2CAP_EVENT_INCOMING_CONNECTION:
			log_info("L2CAP_EVENT_INCOMING_CONNECTION\n");
			l2cap_accept_connection_internal(channel);
			break;
		case L2CAP_EVENT_CHANNEL_OPENED:
	        if (packet[2]) {
	            xputs("control Connection failed\n");
	            return;
		    }
			xputs("control channel connected\n");
			l2cap_create_channel_internal(NULL, l2cap_packet_handler, keyboard, PSM_HID_INTERRUPT, 150);
			//control_channel=channel;
			break;
		case L2CAP_EVENT_CHANNEL_CLOSED:
			log_info("L2CAP_CHANNEL_CLOSED\n");
			//control_channel=0;
			break;
		case L2CAP_EVENT_CREDITS:
			break;
		default:
			log_info("l2cap:unknown(%02x)\n",packet[0]);
			break;
		}	
	}
}

// Bluetooth logic
static void packet_handler (void * connection, uint8_t packet_type, uint16_t channel, uint8_t *packet, uint16_t size)
{
//int i,j;
	bd_addr_t	event_addr;
#if	0
    uint8_t		rfcomm_channel_nr;
    uint8_t		rfcomm_channel_id_temp;
    uint16_t	mtu;
#endif
    
	switch (packet_type) {
		case HCI_EVENT_PACKET:
			switch (packet[0]) {
				case BTSTACK_EVENT_STATE:
					// bt stack activated, get started - set local name
					if (packet[2] == HCI_STATE_WORKING) {
						if(
							!addr_global[0]
						&&	!addr_global[1]
						&&	!addr_global[2]
						&&	!addr_global[3]
						&&	!addr_global[4]
						&&	!addr_global[5]
						){
							//bd address failed -> reset
							Reset();
						}
						xsprintf(lineBuffer,LOCAL_NAME "-%02x%02x%02x%02x%02x%02x",
							addr_global[0],
							addr_global[1],
							addr_global[2],
							addr_global[3],
							addr_global[4],
							addr_global[5]
						);
						log_info("local name:%s\n",lineBuffer);
#if	!__DEBUG
						DelayMs(100);
#endif
                        hci_send_cmd(&hci_write_local_name, lineBuffer);
					}
					break;
				
				case HCI_EVENT_COMMAND_COMPLETE:
					if (COMMAND_COMPLETE_EVENT(packet, hci_read_bd_addr)){
                        bt_flip_addr(event_addr, &packet[6]);
                        log_info("BD-ADDR: %s\n", bd_addr_to_str(event_addr));
                        break;
                    }
					if (COMMAND_COMPLETE_EVENT(packet, hci_write_local_name)){
//						hci_discoverable_control(1);
//						hci_send_cmd(&hci_write_authentication_enable, 1);
						hci_send_cmd(&hci_write_authentication_enable, 0);
                        break;
                    }
					if (COMMAND_COMPLETE_EVENT(packet, hci_write_authentication_enable)){
					inquiry:
#if	!MODE_NO_INQUIRY
						log_info("Inquiry\n");
						hci_send_cmd(&hci_inquiry, HCI_INQUIRY_LAP, INQUIRY_INTERVAL, 0);
#endif
						break;
					}
					if (COMMAND_COMPLETE_EVENT(packet, hci_inquiry_cancel) ) {
						//l2cap_create_channel_internal(NULL, l2cap_packet_handler, keyboard, PSM_HID_INTERRUPT, 150);
						l2cap_create_channel_internal(NULL, l2cap_control_packet_handler, keyboard, PSM_HID_CONTROL, 150);
						break;
					}
                    break;
				case HCI_EVENT_INQUIRY_RESULT:
					
					// ignore further results
					if (haveKeyboard) break;
					
					// ignore none wii remocon
					if(
						!(packet[12]==0x04 && packet[13]==0x25 && packet[14]==0x00)
					&&	!(packet[12]==0x08 && packet[13]==0x05 && packet[14]==0x00)	//RVL-CNT-01-TR
					){
						break;
					}

					// flip addr
					bt_flip_addr(keyboard, &packet[3]);
					
					// show
					log_info("found wii remocon:%s\n",bd_addr_to_str(keyboard));

					haveKeyboard = 1;
					hci_send_cmd(&hci_inquiry_cancel);
					break;
					
				case HCI_EVENT_INQUIRY_COMPLETE:
					log_info("No wii remocon\n");
					goto inquiry;
					break;
				case HCI_EVENT_LINK_KEY_REQUEST:
					// deny link key request
                    log_info("Link key request\n");
                    bt_flip_addr(event_addr, &packet[2]);
					hci_send_cmd(&hci_link_key_request_negative_reply, &event_addr);
					break;
					
				case HCI_EVENT_PIN_CODE_REQUEST:
					// inform about pin code request
                    bt_flip_addr(event_addr, &packet[2]);
					//wii remocon
#if	WII_REMOTE_CONNECTION_SYNC
					bt_flip_addr(packet+2, addr_global);
#endif
                    log_info("Pin code request - using %02x%02x%02x%02x%02x%02x\n",packet[2],packet[3],packet[4],packet[5],packet[6],packet[7]);
					memset(packet+2+6,0x00,10);
					hci_send_cmd(&hci_pin_code_request_reply, &event_addr, 6, &packet[2]);
					break;
	        	case HCI_EVENT_CONNECTION_COMPLETE:
					//log_info("connection complete: status=%d\n",packet[2]);
					if(packet[2]){
						haveKeyboard=0;
						goto inquiry;
					}
					break;
                default:
                    break;
			}
            break;
        default:
            break;
	}
}

static void  heartbeat_handler(struct timer *ts)
{
    static int counter = 0;
	counter++;

	if(wii_remocon_channel_id){
		lineBuffer[0]=0xa2;
		lineBuffer[1]=0x11;
		lineBuffer[2]=0x10<<(counter%4);		//LED Player1
		if(l2cap_send_internal(wii_remocon_channel_id,(uint8_t*)lineBuffer,3)<0){
			//error control init
			wii_button_state=BUTTON_STATE_INIT;
			init_button_state();
		}
	}

    run_loop_set_timer(ts, HEARTBEAT_PERIOD_MS);
    run_loop_add_timer(ts);
}

int main ( void )
{
	RestartWatchdog();
	xdev_out(UART2PutChar);

    // Initialize the processor and peripherals.
    if ( InitializeSystem() != TRUE )
    {
        log_info( "\n\nCould not initialize " LOCAL_NAME " - system.  Halting.\n\n" );
        while (1);
    }
    if ( USBHostInit(0) == TRUE )
    {
        log_info( "\n\n***** " LOCAL_NAME " Initialized *****\n\n" );
    }
    else
    {
        log_info( "\n\nCould not initialize " LOCAL_NAME " - USB.  Halting.\n\n" );
        while (1);
    }

    // Main Processing Loop
    while(!deviceAddress)
    {
        // This demo does not check for overcurrent conditions.  See the
        // USB Host - Data Logger for an example of overcurrent detection
        // with the PIC24F and the USB PICtail Plus.

        // Maintain USB Host State
        USBHostTasks();

        //if (CheckForNewAttach()){
		//	break;
        //}
    }

	init_button_state();
	DelayMs(1000);

	led2_off();

	/// GET STARTED with BTstack ///
	btstack_memory_init();
    run_loop_init(RUN_LOOP_EMBEDDED);
	
    // init HCI
	hci_transport_t    * transport = hci_transport_picusb_instance();
	bt_control_t       * control   = NULL;
    hci_uart_config_t  * config    = NULL;
    remote_device_db_t * remote_db = (remote_device_db_t *) &remote_device_db_memory;
	hci_init(transport, config, control, remote_db);

    // use eHCILL
//    bt_control_cc256x_enable_ehcill(1);
    
    // init L2CAP
    l2cap_init();
    l2cap_register_packet_handler(packet_handler);

#if	MODE_NO_INQUIRY
	l2cap_register_service_internal(NULL, l2cap_control_packet_handler, PSM_HID_CONTROL, 160);
	l2cap_register_service_internal(NULL, l2cap_packet_handler/*l2cap_interrupt_packet_handler*/, PSM_HID_INTERRUPT, 160);
#endif
 
#if	0
	// init RFCOMM
    rfcomm_init();
    rfcomm_register_packet_handler(packet_handler);
    rfcomm_register_service_with_initial_credits_internal(NULL, rfcomm_channel_nr, 100, CREDITS);  // reserved channel, mtu=100, 1 credit

    // init SDP, create record for SPP and register with SDP
    sdp_init();

	memset(spp_service_buffer, 0, sizeof(spp_service_buffer));
    service_record_item_t * service_record_item = (service_record_item_t *) spp_service_buffer;
    sdp_create_spp_service( (uint8_t*) &service_record_item->service_record, 1, "SPP");
    log_info("SDP service buffer size: %u\n", (uint16_t) (sizeof(service_record_item_t) + de_get_len((uint8_t*) &service_record_item->service_record)));
    sdp_register_service_internal(NULL, service_record_item);
#endif

    // set one-shot timer
    timer_source_t heartbeat;
    heartbeat.process = &heartbeat_handler;
    run_loop_set_timer(&heartbeat, HEARTBEAT_PERIOD_MS);
    run_loop_add_timer(&heartbeat);

 	// turn on!
	hci_power_control(HCI_POWER_ON);

//extern linked_list_t db_mem_link_keys;
void load_link_keys(remote_device_db_t * lpremote_device_db);
void print_link_keys(linked_list_t list);
	//load_link_keys(remote_db);
	//print_link_keys(db_mem_link_keys);

    led1_off();

    // go!
    run_loop_execute();	

    return 0;

} // main

#if	ENABLE_KNIGHTRIDER
#define	KR_LEDS			  8
#define	KR_SPWM_WIDTH	 16
#define	KR_CYCLE		128
#define	kr_led_write(x,y)	{if(y){kr_led ## x ## _on();}else{kr_led ## x ## _off();}}
#define	KR_LUM_STEP		  4
#define	KR_LED_PATTERNS	(KR_LEDS*2)
void	knightrider(void)
{
	static uint8_t	led_lum[KR_LEDS];
	static uint8_t	spwm_progress;
	static uint8_t	kr_progress,kr_pat_progress;
	uint8_t			i;

	kr_led_write(1,led_lum[0]>spwm_progress);
	kr_led_write(2,led_lum[1]>spwm_progress);
	kr_led_write(3,led_lum[2]>spwm_progress);
	kr_led_write(4,led_lum[3]>spwm_progress);
	kr_led_write(5,led_lum[4]>spwm_progress);
	kr_led_write(6,led_lum[5]>spwm_progress);
	kr_led_write(7,led_lum[6]>spwm_progress);
	kr_led_write(8,led_lum[7]>spwm_progress);
	spwm_progress=(spwm_progress+1)&(KR_SPWM_WIDTH-1);
	kr_progress=(kr_progress+1)&(KR_CYCLE-1);

	if(!kr_progress){
		for(i=0;i<KR_LEDS;i++){
			led_lum[i]>>=1;
		}
		if(kr_pat_progress<KR_LEDS){
			led_lum[kr_pat_progress]=KR_SPWM_WIDTH;
		}else{
			led_lum[(KR_LEDS-1)-(kr_pat_progress-KR_LEDS)]=KR_SPWM_WIDTH;
		}
		kr_pat_progress=(kr_pat_progress+1)&(KR_LED_PATTERNS-1);
	}
	return;
}
#endif


void hal_tick_init(void)
{
	usb_tick_handler=NULL;
}

void hal_tick_set_handler(void (*tick_handler)(void))
{
	usb_tick_handler=tick_handler;
}

void hal_tick_event(void)
{
	timer_counter++;

	RestartWatchdog();

    if (wii_remocon_channel_id){
        if(timer_counter&0x100){
            led2_off();
        }else{
            led2_on();
        }
    }else{
        led2_on();
    }

	if(led1_on_state && (timer_counter-led1_on_count)>LED1_ON_COUNT){
		led1_off();
		led1_on_state=0;
	}

	// call tick handler
	if(usb_tick_handler){
		usb_tick_handler();
	}

#if	ENABLE_KNIGHTRIDER
	knightrider();
#endif
}
/*************************************************************************
 * EOF main.c
 */

